from google.cloud import bigquery
from typing import List
from google.oauth2 import service_account

from .config import Config
from .exceptions import TooManyTables
from .cli import Colors
from .auth import SA


class Base:
    def __init__(self, id: str, credential_path: str = ""):
        self.config: Config = Config(id=id)
        self.credentials: service_account.Credentials = SA.credentials(
            path=credential_path
        )
        self.bq_client: bigquery.Client = self._client()

    def _client(self) -> bigquery.Client:
        return bigquery.Client(
            project=self.config.project,
            credentials=self.credentials,
            client_options={"scopes": self.config.BQ_CLIENT_SCOPE},
        )


class Table(Base):
    def __init__(self, id: str, credential_path: str = ""):
        super().__init__(id, credential_path)

    @property
    def schema(self):
        table_ref = self.bq_client.dataset(self.config.dataset).table(self.config.table)
        table = self.bq_client.get_table(table_ref)
        return table.schema


class SchemaCompareMixin:
    def difference(self):
        table1 = set(self.resources[0].schema)
        table2 = set(self.resources[1].schema)
        self.diff = table1.symmetric_difference(table2)


class SchemaReprMixin:
    def to_terminal(self):
        repr = [field.to_api_repr() for field in self.diff]
        for i in repr:
            print(
                f"{Colors.OKGREEN}{i['name']}{Colors.ENDC} {Colors.BOLD}- {Colors.ENDC}{Colors.OKCYAN}{i['mode']} {Colors.OKGREEN}{i['type']}{Colors.ENDC}"
            )


class Delta(SchemaReprMixin, SchemaCompareMixin):
    def __init__(self, resources: List[Table]):
        self._resources = self._validate_resources(resources)
        self.difference()

    @property
    def resources(self):
        return self._resources

    def _validate_resources(self, resources: List):
        if len(resources) != 2:
            raise TooManyTables(len(resources))
        return resources


class MakeMigration(Base):
    def __init__(self, config: Config):
        super().__init__(config)



import argparse


class Colors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def add_table1(parser: argparse.ArgumentParser, choices=None, default=None) -> None:
    parser.add_argument(
        "-t1",
        "--table1",
        choices=choices,
        default=default,
        help="Table1 full GBQ id",
    )


def add_table2(parser: argparse.ArgumentParser, choices=None, default=None) -> None:
    parser.add_argument(
        "-t2",
        "--table2",
        choices=choices,
        default=default,
        help="Table2 full GBQ id",
    )


def credentials(parser: argparse.ArgumentParser, choices=None, default=None) -> None:
    parser.add_argument(
        "-c",
        "--credentials",
        choices=choices,
        default=default,
        help="GCP Service Account key",
    )



from .exceptions import MissingConfig, IncorrectID


class Config:
    BQ_CLIENT_SCOPE = ("https://www.googleapis.com/auth/bigquery",)
    _DOT_DELIMITER = "."
    _DATASET_EXCEPTION_CHARS = ["-" "&", "@", "%"]

    def __init__(self, id: str = ""):
        self._gbq_project: str = ""
        self._gbq_dataset: str = ""
        self._gbq_table: str = ""
        self._parse_and_validate_id(id=id)

    @property
    def project(self):
        if not self._gbq_project:
            raise MissingConfig("project")
        return self._gbq_project

    @property
    def dataset(self):
        if not self._gbq_dataset:
            raise MissingConfig("dataset")
        return self._gbq_dataset

    @property
    def table(self):
        if not self._gbq_table:
            raise MissingConfig("table")
        return self._gbq_table

    def _parse_and_validate_id(self, id: str):
        try:
            project, dataset, table = id.split(self._DOT_DELIMITER)
        except ValueError:
            raise IncorrectID(id, "incorrect id delimiter")

        if len(project) > 31:
            raise IncorrectID(project, "incorrect project name lenght")

        if len(dataset) > 1024:
            raise IncorrectID(project, "incorrect dataset name lenght")

        if any(ch in dataset for ch in self._DATASET_EXCEPTION_CHARS):
            raise IncorrectID(project, "invalid dataset name")

        if len(table) > 1024:
            raise IncorrectID(project, "incorrect table name lenght")

        self._gbq_project: str = project
        self._gbq_dataset: str = dataset
        self._gbq_table: str = table



import argparse
import sys
from typing import Optional
from typing import Sequence

from .cli import add_table1, add_table2, credentials
from .version import VERSION
from .gbq import Table, Delta
from .exceptions import MissingConfig


def main(argv: Optional[Sequence[str]] = None) -> int:
    argv = argv if argv is not None else sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("-V", "--version", action="version", version=f"kaisia {VERSION}")

    add_table1(parser)
    add_table2(parser)
    credentials(parser)

    if len(argv) == 0:
        parser.parse_args(["--help"])

    args = parser.parse_args(argv)
    if not args.table1 or not args.table2:
        raise MissingConfig("2 input tables")

    delta = Delta(
        resources=[
            Table(id=args.table1, credential_path=args.credentials),
            Table(id=args.table2, credential_path=args.credentials),
        ]
    )
    delta.to_terminal()


if __name__ == "__main__":
    exit(main())



from google.oauth2 import service_account
import os


class SA:
    @classmethod
    def credentials(self, path: str = "") -> service_account.Credentials:
        try:
            return service_account.Credentials.from_service_account_file(path)
        except (FileNotFoundError, TypeError):
            return service_account.Credentials.from_service_account_file(
                os.environ.get("GOOGLE_APPLICATION_CREDENTIALS")
            )



from .cli import Colors


class MissingConfig(Exception):
    """Exception raised for errors in the config object.

    Attributes:
        param -- configuration parameter
        message -- explanation of the error
    """

    def __init__(self, param, message="Missing config parameter"):
        super().__init__(f"{message}: {Colors.FAIL}{param}")


class IncorrectID(Exception):
    """Exception raised for errors in the gbq parsing of ids.

    Attributes:
        id -- user input id
        message -- explanation of the error
    """

    def __init__(self, id, reason: str = "", message: str = "Parsing faled"):
        super().__init__(
            f"{message} {'with ' + reason if reason else ''}: {Colors.FAIL}{id}"
        )


class TooManyTables(Exception):
    """Exception raised for errors in the amoutn of input tables

    Attributes:
        tables_count -- amount of tables
        message -- explanation of the error
    """

    def __init__(
        self, tables_count, message: str = "Processing faled - too many tables"
    ):
        super().__init__(f"{message}: {Colors.FAIL}{tables_count}")
